"""Python SDK for Victoriabank MIA API"""

import logging

from .victoriabank_mia_sdk import VictoriabankMiaSdk, VictoriabankMiaPaymentException

logger = logging.getLogger(__name__)


class VictoriabankMiaApiRequest:
    """Factory class responsible for creating new instances of the VictoriabankMiaApi class."""

    @staticmethod
    def create(base_url: str = VictoriabankMiaSdk.DEFAULT_BASE_URL):
        """Creates a new instance of VictoriabankMiaApi."""

        client = VictoriabankMiaSdk(base_url=base_url)
        return VictoriabankMiaApi(client)

class VictoriabankMiaApi:
    _client: VictoriabankMiaSdk = None

    REQUIRED_QR_PARAMS = ['header', 'extension']
    REQUIRED_TEST_PAY_PARAMS = ['qrHeaderUUID']

    def __init__(self, client: VictoriabankMiaSdk):
        self._client = client

    #region QR
    def qr_create(self, data: dict, token: str, params: dict = None):
        """CreatePayeeQr - Register new payee-presented QR code

        https://test-ipspj.victoriabank.md/index.html#operations-Qr-post_api_v1_qr"""

        return self._execute_operation(endpoint=VictoriabankMiaSdk.MIA_QR, data=data, token=token, required_params=self.REQUIRED_QR_PARAMS, params=params)

    async def qr_create_async(self, data: dict, token: str, params: dict = None):
        """CreatePayeeQr - Register new payee-presented QR code

        https://test-ipspj.victoriabank.md/index.html#operations-Qr-post_api_v1_qr"""

        return await self._execute_operation_async(endpoint=VictoriabankMiaSdk.MIA_QR, data=data, token=token, required_params=self.REQUIRED_QR_PARAMS, params=params)

    def qr_status(self, qr_id: str, token: str, params: dict = None):
        """getPayeeQrStatus - Get status of payee-presented QR code header, statuses of N last extensions and list of M last payments against each extension

        https://test-ipspj.victoriabank.md/index.html#operations-Qr-get_api_v1_qr__qrHeaderUUID__status"""

        return self._execute_entity_id_operation(endpoint=VictoriabankMiaSdk.MIA_QR_STATUS, entity_id=qr_id, token=token, params=params)

    async def qr_status_async(self, qr_id: str, token: str, params: dict = None):
        """getPayeeQrStatus - Get status of payee-presented QR code header, statuses of N last extensions and list of M last payments against each extension

        https://test-ipspj.victoriabank.md/index.html#operations-Qr-get_api_v1_qr__qrHeaderUUID__status"""

        return await self._execute_entity_id_operation_async(endpoint=VictoriabankMiaSdk.MIA_QR_STATUS, entity_id=qr_id, token=token, params=params)

    def qr_cancel(self, qr_id: str, token: str):
        """CancelPayeeQr-Cancel payee-resented QR code, including active extension, if exists

        https://test-ipspj.victoriabank.md/index.html#operations-Qr-delete_api_v1_qr__qrHeaderUUID_"""

        return self._execute_entity_id_operation(endpoint=VictoriabankMiaSdk.MIA_QR_ID, entity_id=qr_id, token=token, method='DELETE')

    async def qr_cancel_async(self, qr_id: str, token: str):
        """CancelPayeeQr-Cancel payee-resented QR code, including active extension, if exists

        https://test-ipspj.victoriabank.md/index.html#operations-Qr-delete_api_v1_qr__qrHeaderUUID_"""

        return await self._execute_entity_id_operation_async(endpoint=VictoriabankMiaSdk.MIA_QR_ID, entity_id=qr_id, token=token, method='DELETE')
    #endregion

    #region QR Extensions
    def qr_create_extension(self, qr_id: str, data: dict, token: str):
        """CreatePayeeQrExtention - Register new extension for HYBR or STAT payee-presented QR code

        https://test-ipspj.victoriabank.md/index.html#operations-Qr-post_api_v1_qr__qrHeaderUUID__extentions"""

        return self._execute_entity_id_operation(endpoint=VictoriabankMiaSdk.MIA_QR_EXTENSIONS, entity_id=qr_id, token=token, method='POST', data=data)

    async def qr_create_extension_async(self, qr_id: str, data: dict, token: str):
        """CreatePayeeQrExtention - Register new extension for HYBR or STAT payee-presented QR code

        https://test-ipspj.victoriabank.md/index.html#operations-Qr-post_api_v1_qr__qrHeaderUUID__extentions"""

        return await self._execute_entity_id_operation_async(endpoint=VictoriabankMiaSdk.MIA_QR_EXTENSIONS, entity_id=qr_id, token=token, method='POST', data=data)

    def qr_cancel_extension(self, qr_id: str, token: str):
        """CancelHybrExtention - Cancel active extension of hybrid payee-presented QR code

        https://test-ipspj.victoriabank.md/index.html#operations-Qr-delete_api_v1_qr__qrHeaderUUID__active_extension"""

        return self._execute_entity_id_operation(endpoint=VictoriabankMiaSdk.MIA_QR_ACTIVE_EXTENSION, entity_id=qr_id, token=token, method='DELETE')

    async def qr_cancel_extension_async(self, qr_id: str, token: str):
        """CancelHybrExtention - Cancel active extension of hybrid payee-presented QR code

        https://test-ipspj.victoriabank.md/index.html#operations-Qr-delete_api_v1_qr__qrHeaderUUID__active_extension"""

        return await self._execute_entity_id_operation_async(endpoint=VictoriabankMiaSdk.MIA_QR_ACTIVE_EXTENSION, entity_id=qr_id, token=token, method='DELETE')

    def qr_extension_status(self, qr_extension_id: str, token: str,  params: dict = None):
        """getQrExtensionStatus - Get status of QR code extension and list of last N payments against it

        https://test-ipspj.victoriabank.md/index.html#operations-Qr-get_api_v1_qr_extensions__qrExtensionUUID__status"""

        return self._execute_entity_id_operation(endpoint=VictoriabankMiaSdk.MIA_QR_EXTENSION_STATUS, entity_id=qr_extension_id, token=token, params=params)

    async def qr_extension_status_async(self, qr_extension_id: str, token: str,  params: dict = None):
        """getQrExtensionStatus - Get status of QR code extension and list of last N payments against it

        https://test-ipspj.victoriabank.md/index.html#operations-Qr-get_api_v1_qr_extensions__qrExtensionUUID__status"""

        return await self._execute_entity_id_operation_async(endpoint=VictoriabankMiaSdk.MIA_QR_EXTENSION_STATUS, entity_id=qr_extension_id, token=token, params=params)
    #endregion

    #region Transactions
    def transaction_reverse(self, transaction_id: str, token: str):
        """Reverse already processed transaction

        https://test-ipspj.victoriabank.md/index.html#operations-Transaction-delete_api_v1_transaction__id_"""

        return self._execute_entity_id_operation(endpoint=VictoriabankMiaSdk.MIA_TRANSACTION_ID, entity_id=transaction_id, token=token, method='DELETE')

    async def transaction_reverse_async(self, transaction_id: str, token: str):
        """Reverse already processed transaction

        https://test-ipspj.victoriabank.md/index.html#operations-Transaction-delete_api_v1_transaction__id_"""

        return await self._execute_entity_id_operation_async(endpoint=VictoriabankMiaSdk.MIA_TRANSACTION_ID, entity_id=transaction_id, token=token, method='DELETE')

    def transactions_list(self, params: dict, token: str):
        """Transaction list for reconciliation

        https://test-ipspj.victoriabank.md/index.html#operations-Reconciliation-get_api_v1_reconciliation_transactions"""

        return self._execute_operation(endpoint=VictoriabankMiaSdk.MIA_TRANSACTIONS_LIST, data=None, token=token, required_params=None, method='GET', params=params)

    async def transactions_list_async(self, params: dict, token: str):
        """Transaction list for reconciliation

        https://test-ipspj.victoriabank.md/index.html#operations-Reconciliation-get_api_v1_reconciliation_transactions"""

        return await self._execute_operation_async(endpoint=VictoriabankMiaSdk.MIA_TRANSACTIONS_LIST, data=None, token=token, required_params=None, method='GET', params=params)
    #endregion

    #region Signal API
    def signal_get(self, qr_extension_id: str, token: str):
        """https://test-ipspj.victoriabank.md/index.html#operations-Signal-get_api_v1_signal__qrExtensionUUID_"""

        return self._execute_entity_id_operation(endpoint=VictoriabankMiaSdk.MIA_QR_EXTENSION_SIGNAL, entity_id=qr_extension_id, token=token)

    async def signal_get_async(self, qr_extension_id: str, token: str):
        """https://test-ipspj.victoriabank.md/index.html#operations-Signal-get_api_v1_signal__qrExtensionUUID_"""

        return await self._execute_entity_id_operation_async(endpoint=VictoriabankMiaSdk.MIA_QR_EXTENSION_SIGNAL, entity_id=qr_extension_id, token=token)

    #endregion

    #region Payment API
    def test_pay(self, data: dict, token: str):
        """This API enables payment simulation based on the qrHeaderUUID, without requiring any bank or financial institution applications.

        https://test-ipspj-demopay.victoriabank.md/swagger/index.html#operations-Pay-post_api_Pay"""

        return self._execute_operation(endpoint=VictoriabankMiaSdk.TEST_DEMOPAY_URL, data=data, token=token, required_params=self.REQUIRED_TEST_PAY_PARAMS)

    async def test_pay_async(self, data: dict, token: str):
        """This API enables payment simulation based on the qrHeaderUUID, without requiring any bank or financial institution applications.

        https://test-ipspj-demopay.victoriabank.md/swagger/index.html#operations-Pay-post_api_Pay"""

        return await self._execute_operation_async(endpoint=VictoriabankMiaSdk.TEST_DEMOPAY_URL, data=data, token=token, required_params=self.REQUIRED_TEST_PAY_PARAMS)
    #endregion

    #region Health API
    def health_status(self):
        """https://test-ipspj.victoriabank.md/index.html#operations-Health-get_api_v1_health_status"""

        return self._send_request(method='GET', endpoint=VictoriabankMiaSdk.HEALTH_STATUS, token=None)

    async def health_status_async(self):
        """https://test-ipspj.victoriabank.md/index.html#operations-Health-get_api_v1_health_status"""

        return await self._send_request_async(method='GET', endpoint=VictoriabankMiaSdk.HEALTH_STATUS, token=None)
    #endregion

    #region Sync operation
    def _execute_operation(self, endpoint: str, data: dict, token: str, required_params: list, method: str = 'POST', params: dict = None):
        self._validate_params(data=data, required_params=required_params)
        self._validate_access_token(token=token)
        return self._send_request(method=method, endpoint=endpoint, data=data, params=params, token=token)

    def _execute_entity_id_operation(self, endpoint: str, entity_id: str, token: str, method: str = 'GET', data: dict = None, params: dict = None):
        self._validate_id_param(entity_id=entity_id)
        self._validate_access_token(token=token)
        return self._send_request(method=method, endpoint=endpoint, token=token, data=data, params=params, entity_id=entity_id)

    def _send_request(self, method: str, endpoint: str, token: str, data: dict = None, params: dict = None, entity_id: str = None):
        """Sends a request to the specified endpoint."""

        try:
            response = self._client.send_request(method=method, url=endpoint, json_data=data, params=params, token=token, entity_id=entity_id)
        except Exception as ex:
            logger.exception(VictoriabankMiaApi.__qualname__)
            raise VictoriabankMiaPaymentException(f'HTTP error while sending {method} request to endpoint {endpoint}: {ex}') from ex

        return self._client.handle_response(response, endpoint)
    #endregion

    #region Async operation
    async def _execute_operation_async(self, endpoint: str, data: dict, token: str, required_params: list, method: str = 'POST', params: dict = None):
        self._validate_params(data=data, required_params=required_params)
        self._validate_access_token(token=token)
        return await self._send_request_async(method=method, endpoint=endpoint, data=data, params=params, token=token)

    async def _execute_entity_id_operation_async(self, endpoint: str, entity_id: str, token: str, method: str = 'GET', data: dict = None, params: dict = None):
        self._validate_id_param(entity_id=entity_id)
        self._validate_access_token(token=token)
        return await self._send_request_async(method=method, endpoint=endpoint, token=token, data=data, params=params, entity_id=entity_id)

    async def _send_request_async(self, method: str, endpoint: str, token: str, data: dict = None, params: dict = None, entity_id: str = None):
        """Sends async request to the specified endpoint."""

        try:
            response = await self._client.send_request_async(method=method, url=endpoint, json_data=data, params=params, token=token, entity_id=entity_id)
        except Exception as ex:
            logger.exception(VictoriabankMiaApi.__qualname__)
            raise VictoriabankMiaPaymentException(f'HTTP error while sending {method} request to endpoint {endpoint}: {ex}') from ex

        return self._client.handle_response(response, endpoint)
    #endregion

    #region Validation
    @staticmethod
    def _validate_access_token(token: str):
        """Validates the access token."""

        if not token or len(token) == 0:
            raise VictoriabankMiaPaymentException('Access token is not valid. It should be a non-empty string.')

    @staticmethod
    def _validate_id_param(entity_id: str):
        """Validates the ID parameter."""

        if not entity_id:
            raise VictoriabankMiaPaymentException('Missing ID.')

        if len(entity_id) == 0:
            raise VictoriabankMiaPaymentException('Invalid ID parameter. Should be string of 36 characters.')

    @staticmethod
    def _validate_params(data: dict, required_params: list):
        """Validates the parameters."""

        if data and required_params:
            # Check that all required parameters are present
            for param in required_params:
                if data.get(param) is None:
                    raise VictoriabankMiaPaymentException(f'Missing required parameter: {param}')

        return True
    #endregion



"""Python SDK for Victoriabank MIA API"""

import logging

from .victoriabank_mia_sdk import VictoriabankMiaSdk, VictoriabankMiaTokenException, VictoriabankMiaPaymentException
from .victoriabank_mia_auth import VictoriabankMiaAuthRequest, VictoriabankMiaAuth
from .victoriabank_mia_api import VictoriabankMiaApiRequest, VictoriabankMiaApi

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())



"""Python SDK for Victoriabank MIA API"""

import logging
from cryptography.x509 import load_pem_x509_certificate

import httpx
import jwt

logger = logging.getLogger(__name__)


# Based on Python SDK for maib MIA API https://github.com/alexminza/maib-mia-sdk-python (https://pypi.org/project/maib-mia-sdk/)
# IPS Business WebApi https://test-ipspj.victoriabank.md
# IPS DemoPay WebApi https://test-ipspj-demopay.victoriabank.md/swagger/

class VictoriabankMiaSdk:
    # Victoriabank MIA API base urls
    DEFAULT_BASE_URL = 'https://ips-api-pj.vb.md/'
    TEST_BASE_URL = 'https://test-ipspj.victoriabank.md/'
    TEST_DEMOPAY_URL = 'https://test-ipspj-demopay.victoriabank.md/api/pay/'

    HEALTH_STATUS = 'api/v1/health/status'
    AUTH_TOKEN = 'identity/token'

    # Victoriabank MIA API endpoints
    MIA_QR = 'api/v1/qr'
    MIA_QR_ID = 'api/v1/qr/{id}'
    MIA_QR_STATUS = 'api/v1/qr/{id}/status'
    MIA_QR_EXTENSIONS = 'api/v1/qr/{id}/extentions'
    MIA_QR_ACTIVE_EXTENSION = 'api/v1/qr/{id}/active-extension'
    MIA_QR_EXTENSION_STATUS = 'api/v1/qr-extensions/{id}/status'
    MIA_QR_EXTENSION_SIGNAL = 'api/v1/signal/{id}/'

    MIA_TRANSACTION_ID = 'api/v1/transaction/{id}'
    MIA_TRANSACTIONS_LIST = 'api/v1/reconciliation/transactions'

    DEFAULT_TIMEOUT = 30

    _base_url: str = None

    def __init__(self, base_url: str = DEFAULT_BASE_URL):
        self._base_url = base_url

    def _build_url(self, url: str, entity_id: str = None):
        """Build the complete URL for the request"""

        if not url.startswith('https://'):
            url = self._base_url + url

        if entity_id:
            url = url.format(id=entity_id)

        return url

    def _process_response(self, response: httpx.Response):
        if response.is_error:
            logger.error('%s Error: %d %s', VictoriabankMiaSdk.__qualname__, response.status_code, response.text, extra={'method': response.request.method, 'url': response.request.url, 'params': response.request.url.params, 'response_text': response.text, 'status_code': response.status_code})
            #response.raise_for_status()

        if not response.content:
            logger.debug('%s Response: %d', VictoriabankMiaSdk.__qualname__, response.status_code, extra={'response_content': response.content})
            return {}

        response_json: dict = response.json()
        logger.debug('%s Response: %d', VictoriabankMiaSdk.__qualname__, response.status_code, extra={'response_json': response_json})
        return response_json

    def send_request(self, method: str, url: str, form_data: dict = None, json_data: dict = None, params: dict = None, token: str = None, entity_id: str = None):
        """Send a request and parse the response."""

        auth = BearerAuth(token) if token else None
        url = self._build_url(url=url, entity_id=entity_id)

        logger.debug('%s Request: %s %s', VictoriabankMiaSdk.__qualname__, method, url, extra={'method': method, 'url': url, 'form_data': form_data, 'json_data': json_data, 'params': params, 'token': token})
        with httpx.Client() as client:
            response = client.request(method=method, url=url, params=params, data=form_data, json=json_data, auth=auth, timeout=self.DEFAULT_TIMEOUT)
            return self._process_response(response=response)

    async def send_request_async(self, method: str, url: str, form_data: dict = None, json_data: dict = None, params: dict = None, token: str = None, entity_id: str = None):
        """Send async request and parse the response."""

        auth = BearerAuth(token) if token else None
        url = self._build_url(url=url, entity_id=entity_id)

        logger.debug('%s Request: %s %s', VictoriabankMiaSdk.__qualname__, method, url, extra={'method': method, 'url': url, 'form_data': form_data, 'json_data': json_data, 'params': params, 'token': token})
        async with httpx.AsyncClient() as client:
            response = await client.request(method=method, url=url, params=params, data=form_data, json=json_data, auth=auth, timeout=self.DEFAULT_TIMEOUT)
            return self._process_response(response=response)

    @staticmethod
    def handle_response(response: dict, endpoint: str):
        """Handles errors returned by the API."""

        if not isinstance(response, dict):
            raise VictoriabankMiaPaymentException(f"Invalid response received from server for endpoint {endpoint}")

        error_code = response.get('errorCode')
        if error_code:
            error_description = response.get('description')
            raise VictoriabankMiaPaymentException(f'Error sending request to endpoint {endpoint}: {error_description} ({error_code})')

        return response

    @staticmethod
    def decode_callback(callback_jwt: str, public_key_pem: bytes):
        """Decodes and validates the callback data signature."""

        public_key_cert = load_pem_x509_certificate(public_key_pem)
        public_key = public_key_cert.public_key()

        algorithm = 'RS256'
        try:
            decoded_payload: dict = jwt.decode(
                jwt=callback_jwt,
                key=public_key,
                algorithms=[algorithm])

            return decoded_payload
        except Exception as ex:
            raise VictoriabankMiaPaymentException(f'Failed to decode and verify payload signature: {ex}') from ex

#region Auth
class BearerAuth(httpx.Auth):
    """Attaches HTTP Bearer Token Authentication to the given Request object."""
    # https://www.python-httpx.org/advanced/authentication/#custom-authentication-schemes

    token: str = None

    def __init__(self, token: str):
        self.token = token

    def auth_flow(self, request: httpx.Request):
        request.headers['Authorization'] = f'Bearer {self.token}'
        yield request
#endregion

#region Exceptions
class VictoriabankMiaTokenException(Exception):
    pass

class VictoriabankMiaPaymentException(Exception):
    pass
#endregion



"""Python SDK for Victoriabank MIA API"""

import logging

from .victoriabank_mia_sdk import VictoriabankMiaSdk, VictoriabankMiaTokenException

logger = logging.getLogger(__name__)


class VictoriabankMiaAuthRequest:
    """Factory class responsible for creating new instances of the VictoriabankMiaAuth class."""

    @staticmethod
    def create(base_url: str = VictoriabankMiaSdk.DEFAULT_BASE_URL):
        """Creates an instance of the VictoriabankMiaAuth class."""

        client = VictoriabankMiaSdk(base_url=base_url)
        return VictoriabankMiaAuth(client)

class VictoriabankMiaAuth:
    _client: VictoriabankMiaSdk = None

    def __init__(self, client: VictoriabankMiaSdk):
        self._client = client

    #region Generate token API
    def generate_token(self, username: str, password: str):
        """Generate token

        https://test-ipspj.victoriabank.md/index.html#operations-Token-post_identity_token"""

        tokens_data = self._build_generate_data(
            username=username,
            password=password)

        return self._get_tokens(data=tokens_data)

    async def generate_token_async(self, username: str, password: str):
        """Generate token

        https://test-ipspj.victoriabank.md/index.html#operations-Token-post_identity_token"""

        tokens_data = self._build_generate_data(
            username=username,
            password=password)

        return await self._get_tokens_async(data=tokens_data)

    @classmethod
    def _build_generate_data(cls, username: str, password: str):
        if not username and not password:
            raise VictoriabankMiaTokenException('Username and Password are required.')

        tokens_data = {
            'grant_type': 'password',
            'username': username,
            'password': password
        }

        return tokens_data
    #endregion

    #region Refresh token API
    def refresh_token(self, refresh_token: str):
        """Refresh token

        https://test-ipspj.victoriabank.md/index.html#operations-Token-post_identity_token"""

        tokens_data = self._build_refresh_data(refresh_token=refresh_token)
        return self._get_tokens(data=tokens_data)

    async def refresh_token_async(self, refresh_token: str):
        """Refresh token

        https://test-ipspj.victoriabank.md/index.html#operations-Token-post_identity_token"""

        tokens_data = self._build_refresh_data(refresh_token=refresh_token)
        return await self._get_tokens_async(data=tokens_data)

    @classmethod
    def _build_refresh_data(cls, refresh_token: str):
        if not refresh_token:
            raise VictoriabankMiaTokenException('Refresh token is required.')

        tokens_data = {
            'grant_type': 'refresh_token',
            'refresh_token': refresh_token
        }

        return tokens_data
    #endregion

    #region Get tokens
    def _get_tokens(self, data: dict):
        """Get tokens

        https://test-ipspj.victoriabank.md/index.html#operations-Token-post_identity_token"""

        try:
            method = 'POST'
            endpoint = VictoriabankMiaSdk.AUTH_TOKEN
            response = self._client.send_request(method=method, url=endpoint, form_data=data)
        except Exception as ex:
            logger.exception(VictoriabankMiaAuth.__qualname__)
            raise VictoriabankMiaTokenException(f'HTTP error while sending {method} request to endpoint {endpoint}: {ex}') from ex

        result = self._client.handle_response(response, VictoriabankMiaSdk.AUTH_TOKEN)
        return result

    async def _get_tokens_async(self, data: dict):
        """Get tokens

        https://test-ipspj.victoriabank.md/index.html#operations-Token-post_identity_token"""

        try:
            method = 'POST'
            endpoint = VictoriabankMiaSdk.AUTH_TOKEN
            response = await self._client.send_request_async(method=method, url=endpoint, form_data=data)
        except Exception as ex:
            logger.exception(VictoriabankMiaAuth.__qualname__)
            raise VictoriabankMiaTokenException(f'HTTP error while sending {method} request to endpoint {endpoint}: {ex}') from ex

        result = self._client.handle_response(response, VictoriabankMiaSdk.AUTH_TOKEN)
        return result
    #endregion
